//
// Copyright (c) SAS Institute Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package workerrpc

const (
	Ping   = "/ping"
	GetKey = "/getKey"
	Sign   = "/sign"
)

type Request struct {
	KeyName    string
	KeyID      []byte
	Digest     []byte
	Hash       uint
	SaltLength *int
}

type Response struct {
	Value     []byte
	ID        []byte
	Cert      []byte
	Key       string
	Err       string
	Retryable bool
	Usage     bool
}
