//go:build linux || darwin || (windows && amd64)
// +build linux darwin windows,amd64

//
// Copyright (c) SAS Institute Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package passprompt

import (
	"github.com/zalando/go-keyring"
)

var errNotFound = keyring.ErrNotFound

func keyringGet(service, user string) (string, error) {
	return keyring.Get(service, user)
}

func keyringSet(service, user, password string) error {
	return keyring.Set(service, user, password)
}
