//
// Copyright (c) SAS Institute Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package dlog

import (
	"fmt"
	"log"
	"sync/atomic"

	"github.com/kr/pretty"
)

var debugLevel uint32

func SetLevel(level uint32) {
	atomic.StoreUint32(&debugLevel, level)
}

func Printf(level uint32, format string, args ...interface{}) {
	if debugLevel >= level {
		log.Printf("[debug.%d] %s", level, fmt.Sprintf(format, args...))
	}
}

func PrettyPrint(level uint32, v interface{}) {
	Printf(level, "%# v", pretty.Formatter(v))
}
