//
// Copyright (c) SAS Institute Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package config

import (
	"os"
	"path/filepath"
)

func DefaultConfig() string {
	fp := os.Getenv("RELIC_CFG")
	if fp != "" {
		return fp
	}
	profile := os.Getenv("APPDATA")
	if profile != "" {
		// windows
		return filepath.Join(profile, "relic", "relic.yml")
	}
	home := os.Getenv("HOME")
	if home != "" {
		return filepath.Join(home, ".config", "relic", "relic.yml")
	}
	return ""
}
